"""[[DEPRECATED]] Please use citrine.gemtables instead."""

import warnings

warnings.warn("The citrine.ara module is deprecated, please use citrine.gemtables instead",
              DeprecationWarning, stacklevel=2)
